import os
import urllib.request
from abc import ABCMeta, abstractmethod
from io import BytesIO
from typing import Any, Optional
from zipfile import ZipFile

import pandas as pd


class BaseMovieLenstDataLoader(object, metaclass=ABCMeta):
    DOWNLOAD_URL: str
    DEFAULT_PATH: str
    _zf: Optional[ZipFile]

    def __init__(self, zippath: Optional[str] = None, force_download: bool = False):
        if zippath is None:
            zippath = self.DEFAULT_PATH
            if not os.path.exists(zippath):
                if not force_download:
                    download = (
                        input(
                            "Could not find {}.\nCan I download and save it there?[y/N]".format(
                                zippath
                            )
                        ).lower()
                        == "y"
                    )
                else:
                    download = True
                if download:
                    print("start download...")
                    urllib.request.urlretrieve(self.DOWNLOAD_URL, zippath)
                    print("complete")
                else:
                    raise RuntimeError("could not read zipFile")

        self.zf = ZipFile(zippath)

    def _read_as_istream(self, path: str) -> BytesIO:
        bytes_array = self.zf.read(path)
        return BytesIO(bytes_array)

    @abstractmethod
    def read_interaction(self) -> pd.DataFrame:
        raise NotImplementedError("Not implemented")  # pragma: no cover



from abc import ABC, abstractmethod
from io import BytesIO
from typing import IO, Any, Optional, Type

from optuna import Trial, exceptions
from tqdm import tqdm

from ..definitions import InteractionMatrix
from ..evaluator import Evaluator
from .base import BaseRecommender


class TrainerBase(ABC):
    """The trainer class for early-stoppable recommenders.
    The training logic of such recommenders (run-epoch, get-score) will be implemtented in the trainers."""

    def __init__(self, **kwargs: Any) -> None:
        pass

    @abstractmethod
    def load_state(self, ifs: IO) -> None:
        """Load past state (with best validation score).

        Args:
            ifs (IO): The stream from which past state will be restored.
        """
        raise NotImplementedError()  # pragma: no cover

    @abstractmethod
    def save_state(self, ofs: IO) -> None:
        """Save the current state into stream.

        Args:
            ofs (IO): the stream to which current state will be saved.
        """
        raise NotImplementedError()  # pragma: no cover

    @abstractmethod
    def run_epoch(self) -> None:
        """Run single epoch."""
        raise NotImplementedError()  # pragma: no cover


class BaseRecommenderWithEarlyStopping(BaseRecommender):
    """The base class for all the early-stoppable recommenders.

    Args:
        X_train_all (csr_matrix|csc_matrix): The train interaction matrix.
        max_epoch (int, optional): The maximal number of epochs to be run. Defaults to 512.
        validate_epoch (int, optional): Frequency of validation score measurement (if any). Defaults to 5.
        score_degradation_max (int, optional): Maximal number of allowed score degradation. Defaults to 5.
    """

    trainer_class: Type[TrainerBase]

    def __init__(
        self,
        X_train_all: InteractionMatrix,
        max_epoch: int = 512,
        validate_epoch: int = 5,
        score_degradation_max: int = 5,
        **kwargs: Any,
    ):

        super().__init__(X_train_all, **kwargs)
        self.max_epoch = max_epoch
        self.validate_epoch = validate_epoch
        if max_epoch < validate_epoch:
            raise ValueError("max_epoch must be greater than validate_epoch.")
        self.score_degradation_max = score_degradation_max
        self.trainer: Optional[TrainerBase] = None
        self.best_state: Optional[bytes] = None

    @abstractmethod
    def _create_trainer(self) -> TrainerBase:
        pass  # pragma: no cover

    def start_learning(self) -> None:
        self.trainer = self._create_trainer()

    def run_epoch(self) -> None:
        if self.trainer is None:
            raise RuntimeError("'run_epoch' called before initializing the trainer.")
        self.trainer.run_epoch()

    def save_state(self) -> None:
        if self.trainer is None:
            raise RuntimeError("'save_state' called before initializing the trainer.")
        with BytesIO() as ofs:
            self.trainer.save_state(ofs)
            self.best_state = ofs.getvalue()

    def load_state(self) -> None:
        if self.trainer is None:
            raise RuntimeError("'load_state' called before initializing the trainer.")
        if self.best_state is None:
            raise RuntimeError("'load_state' called before achieving any results.")
        with BytesIO(self.best_state) as ifs:
            self.trainer.load_state(ifs)

    def _learn(self) -> None:
        self.learn_with_optimizer(None, None)

    def learn_with_optimizer(
        self, evaluator: Optional[Evaluator], trial: Optional[Trial]
    ) -> None:
        self.start_learning()
        best_score = -float("inf")
        n_score_degradation = 0

        with tqdm(total=self.max_epoch) as progress_bar:
            for epoch in range(self.max_epoch):
                self.run_epoch()
                progress_bar.update(1)
                if (epoch + 1) % self.validate_epoch:
                    continue

                if evaluator is None:
                    continue

                target_score = evaluator.get_target_score(self)

                progress_bar.set_description(f"valid_score={target_score}")
                relevant_score = target_score
                if relevant_score > best_score:
                    best_score = relevant_score
                    self.save_state()
                    self.learnt_config["max_epoch"] = epoch + 1
                    n_score_degradation = 0
                else:
                    n_score_degradation += 1
                    if n_score_degradation >= self.score_degradation_max:
                        break
                if trial is not None:
                    trial.report(-relevant_score, epoch)
                    if trial.should_prune():
                        raise exceptions.TrialPruned()

            if evaluator is not None:
                self.load_state()



import numpy as np
import pytest
import scipy.sparse as sps

from irspack.recommenders import TopPopRecommender

X = np.random.rand(200, 512)
X[X <= 0.9] = 0
X[X > 0.9] = 1
X = sps.csr_matrix(X)


def test_toppop() -> None:
    rec = TopPopRecommender(X)
    with pytest.raises(RuntimeError):
        _ = rec.get_score_cold_user_remove_seen(X)
    rec.learn()
    score = rec.get_score_cold_user(X)
    assert score.shape == X.shape
    np.testing.assert_allclose(score[0, :], score[-1, :])

    score_hot = rec.get_score_remove_seen(np.arange(X.shape[0]))
    assert np.all(np.isinf(score_hot[X.nonzero()]))



from typing import Optional

import numpy as np
import pytest
import scipy.sparse as sps

from irspack.dataset.movielens import MovieLens100KDataManager
from irspack.split import (
    UserTrainTestInteractionPair,
    split_dataframe_partial_user_holdout,
)

RNS = np.random.RandomState(0)

df = MovieLens100KDataManager(force_download=True).read_interaction()

test_configs = [
    (None, None, 0.1, 0.15),
    (None, 3, 0.2, 0.25),
    (4, None, 0.3, 0.35),
    (5, 6, 0.0, 1.0),
]
test_configs_with_time = [
    config + (tsc,) for tsc in ["timestamp", None] for config in test_configs
]


@pytest.mark.parametrize(
    "n_val_user, n_test_user, val_user_ratio, test_user_ratio, time_colname",
    test_configs_with_time,
)
def test_user_level_split(
    n_val_user: Optional[int],
    n_test_user: Optional[int],
    val_user_ratio: float,
    test_user_ratio: float,
    time_colname: Optional[str],
) -> None:
    n_users_all = len(set(df.userId))
    dataset, mid_list = split_dataframe_partial_user_holdout(
        df,
        user_column="userId",
        item_column="movieId",
        time_column=time_colname,
        val_user_ratio=val_user_ratio,
        test_user_ratio=test_user_ratio,
        n_val_user=n_val_user,
        n_test_user=n_test_user,
        heldout_ratio_val=0.3,
        heldout_ratio_test=0.5,
    )
    assert len(mid_list) == len(set(df.movieId))
    train = dataset["train"]
    train_invalid = UserTrainTestInteractionPair(
        train.user_ids, train.X_train[:, :-1], None
    )
    with pytest.raises(ValueError):
        train_invalid.concat(train)
    with pytest.raises(ValueError):
        _ = UserTrainTestInteractionPair(
            train.user_ids, train.X_train, train.X_train[1:]
        )

    with pytest.raises(ValueError):
        _ = UserTrainTestInteractionPair(
            train.user_ids, train.X_train, train.X_train, mid_list[:-1]
        )

    def get_n_right_answer(ratio: float, n: Optional[int]) -> int:
        if n is not None:
            return n
        else:
            return int(n_users_all * ratio)

    val = dataset["val"]
    assert val.n_users == get_n_right_answer(val_user_ratio, n_val_user)
    test = dataset["test"]
    assert test.n_users == get_n_right_answer(test_user_ratio, n_test_user)

    if time_colname is not None:
        for d in [val, test]:
            _df_train = d.df_train().merge(
                df[["userId", "movieId", "timestamp"]].rename(
                    columns={"userId": "user_id", "movieId": "item_id"}
                )
            )
            _df_test = d.df_test().merge(
                df[["userId", "movieId", "timestamp"]].rename(
                    columns={"userId": "user_id", "movieId": "item_id"}
                )
            )
            _train_max_time = _df_train.groupby("user_id").timestamp.max()
            _test_min_time = _df_test.groupby("user_id").timestamp.min()
            common_index = np.intersect1d(_train_max_time.index, _test_min_time.index)
            assert common_index.shape[0] > 0
            assert np.all(
                _train_max_time.reindex(common_index)
                <= _test_min_time.reindex(common_index)
            )

    assert train.X_test.count_nonzero() == 0
    train_val = train.concat(val)
    assert train_val.X_test[: train.n_users].count_nonzero() == 0
    assert (train_val.X_test[train.n_users :] - val.X_test).count_nonzero() == 0

    assert (
        train_val.X_train - sps.vstack([train.X_all, val.X_train])
    ).count_nonzero() == 0

    for user_data, ratio in [(val, 0.3), (test, 0.5)]:
        X_learn = user_data.X_train
        X_predict = user_data.X_test
        assert X_predict is not None
        intersect = X_learn.multiply(X_predict)
        assert intersect.count_nonzero() == 0
        index = RNS.choice(np.arange(user_data.n_users), size=10)
        for i in index:
            nnz_learn = X_learn[i].nonzero()[1].shape[0]
            nnz_predict = X_predict[i].nonzero()[1].shape[0]
            assert ratio >= (nnz_predict - 1) / (nnz_learn + nnz_predict)
            assert ratio <= (nnz_predict + 1) / (nnz_learn + nnz_predict)


def test_user_level_split_val_fixed_n() -> None:
    ratio = 0.3
    dataset, mid_list = split_dataframe_partial_user_holdout(
        df,
        user_column="userId",
        item_column="movieId",
        n_test_user=30,
        n_val_user=30,
        n_heldout_val=1,
        heldout_ratio_test=ratio,
    )
    assert len(mid_list) == len(set(df.movieId))

    train = dataset["train"]
    train_invalid = UserTrainTestInteractionPair(
        train.user_ids, train.X_train[:, :-1], None
    )
    with pytest.raises(ValueError):
        train_invalid.concat(train)
    with pytest.raises(ValueError):
        _ = UserTrainTestInteractionPair(
            train.user_ids, train.X_train, train.X_train[1:]
        )

    val = dataset["val"]
    test = dataset["test"]
    assert train.X_test.count_nonzero() == 0
    train_val = train.concat(val)
    assert train_val.X_test[: train.n_users].count_nonzero() == 0
    assert (train_val.X_test[train.n_users :] - val.X_test).count_nonzero() == 0

    assert (
        train_val.X_train - sps.vstack([train.X_all, val.X_train])
    ).count_nonzero() == 0

    val_X_test = val.X_test
    assert np.all(val_X_test.sum(axis=1).A1 <= 1)

    X_learn = test.X_train
    X_predict = test.X_test
    assert X_predict is not None
    intersect = X_learn.multiply(X_predict)
    assert intersect.count_nonzero() == 0
    index = RNS.choice(np.arange(test.n_users), size=10)
    for i in index:
        nnz_learn = X_learn[i].nonzero()[1].shape[0]
        nnz_predict = X_predict[i].nonzero()[1].shape[0]
        assert ratio >= (nnz_predict - 1) / (nnz_learn + nnz_predict)
        assert ratio <= (nnz_predict + 1) / (nnz_learn + nnz_predict)



from typing import List

from irspack.parameter_tuning import (
    CategoricalSuggestion,
    IntegerSuggestion,
    LogUniformSuggestion,
    Suggestion,
)
from irspack.user_cold_start.recommenders.cb_knn import UserCBCosineKNNRecommender
from irspack.user_cold_start.recommenders.linear import LinearMethodRecommender
from irspack.user_cold_start.recommenders.popular import TopPopularRecommender

from .base_optimizer import BaseOptimizer


class TopPopularOptimizer(BaseOptimizer):
    recommender_class = TopPopularRecommender
    default_tune_range: List[Suggestion] = []


class LinearMethodOptimizer(BaseOptimizer):
    recommender_class = LinearMethodRecommender
    default_tune_range = [
        LogUniformSuggestion("reg", 1e-1, 1e4),
        CategoricalSuggestion("fit_intercept", [True, False]),
    ]


class UserCBCosineKNNOptimizer(BaseOptimizer):
    recommender_class = UserCBCosineKNNRecommender
    default_tune_range = [
        IntegerSuggestion("top_k", 5, 2000),
        LogUniformSuggestion("shrink", 1e-2, 1e2),
    ]


__all__ = ["TopPopularOptimizer", "LinearMethodOptimizer", "UserCBCosineKNNOptimizer"]



from collections import defaultdict
from typing import List

import numpy as np
import pytest
import scipy.sparse as sps
from sklearn.metrics import ndcg_score

from irspack.evaluator import Evaluator
from irspack.recommenders.base import BaseRecommender


class MockRecommender(BaseRecommender):
    def __init__(self, X_all: sps.csr_matrix, scores: np.ndarray) -> None:
        super().__init__(X_all)
        self.scores = scores

    def get_score(self, user_indices: np.ndarray) -> np.ndarray:
        return self.scores[user_indices]

    def _learn(self) -> None:
        pass


@pytest.mark.parametrize("U, I, R", [(10, 30, 10), (100, 30, 5), (30, 100, 2)])
def test_restriction_global(U: int, I: int, R: int) -> None:
    rns = np.random.RandomState(42)
    recommendable = rns.choice(np.arange(I), replace=False, size=R)
    scores = rns.randn(U, I)
    X_gt = (rns.rand(U, I) >= 0.3).astype(np.float64)
    eval = Evaluator(
        sps.csr_matrix(X_gt),
        offset=0,
        cutoff=I,
        n_threads=1,
        recommendable_items=recommendable,
    )
    # empty mask
    mock_rec = MockRecommender(sps.csr_matrix(X_gt.shape), scores)
    my_score = eval.get_score(mock_rec)
    sklearn_metrics = defaultdict(list)
    for i in range(scores.shape[0]):
        if X_gt[i, recommendable].sum() == 0:
            continue
        ndcg = ndcg_score(
            X_gt[i, recommendable][None, :], scores[i, recommendable][None, :]
        )
        sklearn_metrics["ndcg"].append(ndcg)

    assert my_score["ndcg"] == pytest.approx(np.mean(sklearn_metrics["ndcg"]), abs=1e-8)


@pytest.mark.parametrize("U, I", [(10, 30), (100, 30), (30, 100)])
def test_restriction_local(U: int, I: int) -> None:
    rns = np.random.RandomState(42)
    recommendables: List[np.ndarray] = []
    for _ in range(U):
        recommendables.append(
            rns.choice(np.arange(I), replace=False, size=rns.randint(2, I))
        )
    scores = rns.randn(U, I)
    X_gt = (rns.rand(U, I) >= 0.3).astype(np.float64)
    eval = Evaluator(
        sps.csr_matrix(X_gt),
        offset=0,
        cutoff=I,
        n_threads=1,
        per_user_recommendable_items=recommendables,
    )
    # empty mask
    mock_rec = MockRecommender(sps.csr_matrix(X_gt.shape), scores)
    my_score = eval.get_score(mock_rec)
    sklearn_metrics = defaultdict(list)
    for i in range(scores.shape[0]):
        if X_gt[i, recommendables[i]].sum() == 0:
            continue
        ndcg = ndcg_score(
            X_gt[i, recommendables[i]][None, :], scores[i, recommendables[i]][None, :]
        )
        sklearn_metrics["ndcg"].append(ndcg)

    assert my_score["ndcg"] == pytest.approx(np.mean(sklearn_metrics["ndcg"]), abs=1e-8)


@pytest.mark.parametrize("U, I", [(10, 30)])
def test_irregular(U: int, I: int) -> None:
    rns = np.random.RandomState(42)
    recommendables: List[np.ndarray] = []
    X_gt = (rns.rand(U, I) >= 0.3).astype(np.float64)
    _ = Evaluator(
        sps.csr_matrix(X_gt),
        offset=0,
        cutoff=I,
        n_threads=1,
        per_user_recommendable_items=[],
    )
    _ = Evaluator(
        sps.csr_matrix(X_gt),
        offset=0,
        cutoff=I,
        n_threads=1,
        per_user_recommendable_items=[[0]],
    )
    _ = Evaluator(
        sps.csr_matrix(X_gt),
        offset=0,
        cutoff=I,
        n_threads=1,
        per_user_recommendable_items=[[0] for _ in range(X_gt.shape[0])],
    )

    with pytest.raises(ValueError):
        _ = Evaluator(
            sps.csr_matrix(X_gt),
            offset=0,
            cutoff=I,
            n_threads=1,
            per_user_recommendable_items=[[0], [0]],
        )
    with pytest.raises(ValueError):
        eval = Evaluator(
            sps.csr_matrix(X_gt),
            offset=0,
            cutoff=I,
            n_threads=1,
            per_user_recommendable_items=[[0, 0]],
        )
    with pytest.raises(ValueError):
        eval = Evaluator(
            sps.csr_matrix(X_gt),
            offset=0,
            cutoff=I,
            n_threads=1,
            per_user_recommendable_items=[
                [
                    I,
                ]
            ],
        )
    # empty mask



from collections import defaultdict

import numpy as np
import pytest
import scipy.sparse as sps
from sklearn.metrics import average_precision_score, ndcg_score

from irspack.evaluator import Evaluator, EvaluatorWithColdUser
from irspack.recommenders import P3alphaRecommender
from irspack.recommenders.base import BaseRecommender
from irspack.split import rowwise_train_test_split


class MockRecommender(BaseRecommender):
    def __init__(self, X_all: sps.csr_matrix, scores: np.ndarray) -> None:
        super().__init__(X_all)
        self.scores = scores

    def get_score(self, user_indices: np.ndarray) -> np.ndarray:
        return self.scores[user_indices]

    def _learn(self) -> None:
        pass


@pytest.mark.parametrize("U, I", [(10, 5), (10, 30), (3000, 5)])
def test_metrics(U: int, I: int) -> None:
    rns = np.random.RandomState(42)
    scores = rns.randn(U, I)
    X_gt = (rns.rand(U, I) >= 0.3).astype(np.float64)
    eval = Evaluator(sps.csr_matrix(X_gt), offset=0, cutoff=I, n_threads=4)
    # empty mask
    mock_rec = MockRecommender(sps.csr_matrix(X_gt.shape), scores)
    my_score = eval.get_score(mock_rec)
    sklearn_metrics = defaultdict(list)
    for i in range(scores.shape[0]):
        if X_gt[i].sum() == 0:
            continue
        sklearn_metrics["map"].append(average_precision_score(X_gt[i], scores[i]))
        sklearn_metrics["ndcg"].append(ndcg_score(X_gt[i][None, :], scores[i][None, :]))

    for key in ["map", "ndcg"]:
        assert my_score[key] == pytest.approx(np.mean(sklearn_metrics[key]), abs=1e-8)


@pytest.mark.parametrize("U, I, C", [(10, 5, 5), (10, 30, 29)])
def test_metrics_with_cutoff(U: int, I: int, C: int) -> None:
    rns = np.random.RandomState(42)
    scores = rns.randn(U, I)
    X_gt = (rns.rand(U, I) >= 0.3).astype(np.float64)
    eval = Evaluator(sps.csr_matrix(X_gt), offset=0, cutoff=C, n_threads=2)
    eval_finer_chunk = Evaluator(
        sps.csr_matrix(X_gt), offset=0, cutoff=C, n_threads=2, mb_size=1
    )
    # empty mask
    mock_rec = MockRecommender(sps.csr_matrix(X_gt.shape), scores)
    my_score = eval.get_score(mock_rec)
    my_score_finer = eval_finer_chunk.get_score(mock_rec)
    for key in my_score:
        assert my_score_finer[key] == pytest.approx(my_score[key])

    ndcg = 0.0
    valid_users = 0
    map = 0.0
    precision = 0.0
    recall = 0.0
    item_appearance_count = np.zeros((I,), dtype=np.float64)
    for i in range(U):
        nzs = set(X_gt[i].nonzero()[0])
        if len(nzs) == 0:
            continue
        valid_users += 1
        ndcg += ndcg_score(X_gt[[i]], scores[[i]], k=C)
        recommended = scores[i].argsort()[::-1][:C]
        recall_denom = min(C, len(nzs))
        ap = 0.0
        current_hit = 0
        for i, rec in enumerate(recommended):
            item_appearance_count[rec] += 1.0
            if rec in nzs:
                current_hit += 1
                ap += current_hit / float(i + 1)
        ap /= recall_denom
        map += ap
        recall += current_hit / recall_denom
        precision += current_hit / C
    entropy = (lambda p: -p.dot(np.log(p)))(
        item_appearance_count / item_appearance_count.sum()
    )
    item_appearance_sorted_normalized = (
        np.sort(item_appearance_count) / item_appearance_count.sum()
    )
    lorentz_curve = np.cumsum(item_appearance_sorted_normalized)

    gini_index = 0
    delta = 1 / I
    for i in range(I):
        f = 2 * (((i + 1) / I) - lorentz_curve[i])
        gini_index += delta * f

    assert my_score["ndcg"] == pytest.approx(ndcg / valid_users)
    assert my_score["map"] == pytest.approx(map / valid_users, abs=1e-8)
    assert my_score["precision"] == pytest.approx(precision / valid_users, abs=1e-8)
    assert my_score["recall"] == pytest.approx(recall / valid_users, abs=1e-8)
    assert my_score["entropy"] == pytest.approx(entropy)
    assert my_score["gini_index"] == pytest.approx(gini_index)


@pytest.mark.parametrize("U, I, U_test", [(10, 5, 3), (10, 30, 8)])
def test_metrics_ColdUser(U: int, I: int, U_test: int) -> None:
    rns = np.random.RandomState(42)
    uvec = rns.randn(U + U_test, 3)
    ivec = rns.randn(I, 3)
    true_score = uvec.dot(ivec.T)  # + rns.randn(U, I)
    X = sps.csr_matrix((true_score > 0).astype(np.float64))
    X_train = X[:U]
    X_val = X[U:]
    X_val_learn, X_val_target = rowwise_train_test_split(X_val, random_seed=0)
    X_train_all = sps.vstack([X_train, X_val_learn])
    hot_evaluator = Evaluator(
        sps.csr_matrix(X_val_target), offset=U, cutoff=I // 2, n_threads=2
    )

    rec = P3alphaRecommender(X_train_all)
    rec.learn()
    hot_score = hot_evaluator.get_score(rec)
    with pytest.warns(UserWarning):
        cold_evaluator = EvaluatorWithColdUser(
            X_val_learn.tocsc(), X_val_target, cutoff=I // 2, mb_size=5
        )  # csc matrix input should raise warning about
        # memory ordering, as csc-csc matrix product will be csc,
        # hence col-major matrix when made dense.
        cold_score = cold_evaluator.get_score(rec)

    shuffle_index = np.arange(X_val_learn.shape[0])
    rns.shuffle(shuffle_index)
    cold_evaluator_shuffled = EvaluatorWithColdUser(
        X_val_learn[shuffle_index], X_val_target[shuffle_index], cutoff=I // 2
    )
    cold_score_shuffled = cold_evaluator_shuffled.get_score(rec)
    for key in cold_score:
        assert cold_score_shuffled[key] == pytest.approx(cold_score[key])

    for key in hot_score:
        assert hot_score[key] == pytest.approx(cold_score[key], abs=1e-8)


@pytest.mark.parametrize("U, I, C", [(10, 5, 5), (10, 30, 29)])
def test_recommender_check(U: int, I: int, C: int) -> None:
    rns = np.random.RandomState(42)
    scores = rns.randn(U, I)
    X_gt = (rns.rand(U, I) >= 0.3).astype(np.float64)
    eval = Evaluator(sps.csr_matrix(X_gt), offset=0, cutoff=C, n_threads=2)
    mock_rec_too_few_users = MockRecommender(sps.csr_matrix((U - 1, I)), scores[1:])
    with pytest.raises(ValueError):
        eval.get_score(mock_rec_too_few_users)
    mock_rec_too_few_items = MockRecommender(sps.csr_matrix((U, I - 1)), scores[:, 1:])
    with pytest.raises(ValueError):
        eval.get_score(mock_rec_too_few_items)
    moc_rec_valid = MockRecommender(sps.csr_matrix((U, I)), scores)
    eval.get_score(moc_rec_valid)



import numpy as np
import pytest
import scipy.sparse as sps

from irspack.definitions import InteractionMatrix, ProfileMatrix
from irspack.evaluator import Evaluator
from irspack.split import rowwise_train_test_split

RNS = np.random.RandomState(0)

profile = sps.csr_matrix(RNS.rand(3, 10) >= 0.7).astype(np.float64)
X_cf = sps.csr_matrix(RNS.rand(3, 30) >= 0.7).astype(np.float64)

profile = sps.vstack([profile for _ in range(10)])  # so many duplicates!
X_cf = sps.vstack([X_cf for _ in range(10)])


@pytest.mark.parametrize("X, profile", [(X_cf, profile)])
def test_cb2cf(X: InteractionMatrix, profile: ProfileMatrix) -> None:

    """Fit IALS & let mlp overfit.

    Args:
        X (InteractionMatrix): user_item interaction matrix
        profile (ProfileMatrix): profile
    """
    try:
        from irspack.user_cold_start.cb2cf import CB2IALSOptimizer
    except:
        pytest.skip("Failed to import jax.")
        raise

    X_cf_train_all, X_val = rowwise_train_test_split(
        X_cf, test_ratio=0.5, random_seed=0
    )
    evaluator = Evaluator(X_val, 0)
    optim = CB2IALSOptimizer(
        X_cf_train_all,
        evaluator,
        profile,
    )
    cb2cfrec, t, mlp_config = optim.search_all(
        20,
        cf_fixed_params=dict(n_components=5, alpha=0, reg=1e-3, max_cg_steps=30),
        random_seed=0,
    )
    vec_reconstruction = cb2cfrec.mlp.predict(profile.astype(np.float32).toarray())
    vec_target = cb2cfrec.cf_rec.get_user_embedding()

    residual = ((vec_reconstruction - vec_target) ** 2).sum() / (vec_target ** 2).sum()
    assert residual <= 1e-1



from ..utils import rowwise_train_test_split
from .specified import holdout_specific_interactions
from .time import split_last_n_interaction_df
from .userwise import UserTrainTestInteractionPair, split_dataframe_partial_user_holdout

__all__ = [
    "UserTrainTestInteractionPair",
    "split_dataframe_partial_user_holdout",
    "holdout_specific_interactions",
    "rowwise_train_test_split",
    "split_last_n_interaction_df",
]



# __version__ = "0.1.0.dev3"
from pkg_resources import DistributionNotFound, get_distribution

try:
    __version__ = get_distribution("irspack").version
except DistributionNotFound:
    # package is not installed
    pass



from typing import Any, Dict, List, Optional, Tuple

import numpy as np
import scipy.sparse as sps

from irspack.definitions import DenseScoreArray, InteractionMatrix, UserIndexArray
from irspack.recommenders import BaseRecommender


class IDMappedRecommender:
    """A utility class that helps  mapping user/item ids to index, retrieving recommendation score,
    and making a recommendation.

    Args:
        recommender:
            The backend base recommender which transforms user/item ids.
        user_ids:
            user_ids which correspods to the rows of ``recommender.X_train_all``.
        item_ids:
            item_ids which correspods to the columns of ``recommender.X_train_all``.

    Raises:
        ValueError: When recommender and user_ids/item_ids are inconsistent.
        ValueError: When there is a duplicate in user_ids.
        ValueError: When there is a duplicate in item_ids.

    """

    def __init__(
        self, recommender: BaseRecommender, user_ids: List[Any], item_ids: List[Any]
    ):

        if (recommender.n_users != len(user_ids)) or (
            recommender.n_items != len(item_ids)
        ):
            raise ValueError(
                "The recommender and user/item ids have inconsistent lengths."
            )

        self.recommender = recommender
        self.user_ids = user_ids
        self.item_ids = item_ids
        self.user_id_to_index = {user_id: i for i, user_id in enumerate(user_ids)}
        self.item_id_to_index = {item_id: i for i, item_id in enumerate(item_ids)}

    def _item_id_list_to_array(self, ids: List[Any]) -> np.ndarray:
        return np.sort(
            np.asarray(
                [
                    self.item_id_to_index[id]
                    for id in ids
                    if id in self.item_id_to_index
                ],
                dtype=np.int64,
            )
        )

    def _score_to_recommended_items(
        self,
        score: DenseScoreArray,
        cutoff: int,
        allowed_item_ids: Optional[List[Any]] = None,
        forbidden_item_ids: Optional[List[Any]] = None,
    ) -> List[Tuple[Any, float]]:
        if allowed_item_ids is not None:
            allowed_item_indices = self._item_id_list_to_array(allowed_item_ids)
            high_score_inds = allowed_item_indices[
                score[allowed_item_indices].argsort()[::-1]
            ]
        else:
            high_score_inds = score.argsort()[::-1]
        recommendations: List[Tuple[Any, float]] = []
        for i in high_score_inds:
            i_int = int(i)
            score_this = score[i_int]
            item_id = self.item_ids[i_int]
            if np.isinf(score_this):
                continue
            if forbidden_item_ids is not None:
                if item_id in forbidden_item_ids:
                    continue
            recommendations.append((item_id, float(score_this)))
            if len(recommendations) >= cutoff:
                break
        return recommendations

    def get_recommendation_for_known_user_id(
        self,
        user_id: Any,
        cutoff: int = 20,
        allowed_item_ids: Optional[List[Any]] = None,
        forbidden_item_ids: Optional[List[Any]] = None,
    ) -> List[Tuple[Any, float]]:
        if user_id not in self.user_ids:
            raise RuntimeError(f"User with user_id {user_id} not found.")
        user_index: UserIndexArray = np.asarray(
            [self.user_id_to_index[user_id]], dtype=np.int64
        )

        score = self.recommender.get_score_remove_seen(user_index)[0, :]
        return self._score_to_recommended_items(
            score,
            cutoff=cutoff,
            allowed_item_ids=allowed_item_ids,
            forbidden_item_ids=forbidden_item_ids,
        )

    def get_recommendation_for_new_user(
        self,
        item_ids: List[Any],
        cutoff: int = 20,
        allowed_item_ids: Optional[List[Any]] = None,
        forbidden_item_ids: Optional[List[Any]] = None,
    ) -> List[Tuple[Any, float]]:
        cols = self._item_id_list_to_array(item_ids)
        rows = np.zeros_like(cols)
        data = np.ones(cols.shape, dtype=np.float64)
        X_input = sps.csr_matrix((data, (rows, cols)), shape=(1, len(self.item_ids)))
        score = self.recommender.get_score_cold_user_remove_seen(X_input)[0]
        return self._score_to_recommended_items(
            score,
            cutoff,
            allowed_item_ids=allowed_item_ids,
            forbidden_item_ids=forbidden_item_ids,
        )



from typing import Any, Dict, List, Optional, Tuple

import numpy as np
import pandas as pd
from scipy import sparse as sps
from sklearn.model_selection import train_test_split

from irspack.definitions import InteractionMatrix
from irspack.split.time import split_last_n_interaction_df
from irspack.utils import rowwise_train_test_split


class UserTrainTestInteractionPair:
    """A class to hold users' train & test (if any) interactions and their ids.

    Args:
        user_ids:
            List of user ids. Its ``i``-th element should correspond to ``i``-th row
            of ``X_train``.
        X_train:
            The train part of interactions.
        X_test:
            The test part of interactions (if any).
            If ``None``, an empty matrix with the shape of ``X_train``
            will be created. Defaults to None.

    Raises:
        ValueError:
            when ``X_train`` and ``user_ids`` have inconsistent size.
        ValueError:
            when ``X_train`` and ``X_test`` have inconsistent size.

    """

    X_train: sps.csr_matrix
    """The train part of users' interactions."""
    X_test: sps.csr_matrix
    """The test part of users' interactions."""
    n_users: int
    """The number of users"""
    n_items: int
    """The number of items"""
    X_all: sps.csr_matrix
    """If ``X_test`` is not ``None``, equal to ``X_train + X_test``.Otherwise equals X_train."""

    def __init__(
        self,
        user_ids: List[Any],
        X_train: InteractionMatrix,
        X_test: Optional[InteractionMatrix],
        item_ids: Optional[List[Any]] = None,
    ):

        if len(user_ids) != X_train.shape[0]:
            raise ValueError("user_ids and X_train have different shapes.")

        if X_test is not None:
            if X_train.shape != X_test.shape:
                raise ValueError("X_train and X_test have different shapes.")
            X_test = sps.csr_matrix(X_test)
        else:
            X_test = sps.csr_matrix(X_train.shape, dtype=X_train.dtype)
        self.user_ids = [x for x in user_ids]
        self.X_train = sps.csr_matrix(X_train)
        self.X_test = X_test
        self.n_users = self.X_train.shape[0]
        self.n_items = self.X_train.shape[1]
        self.X_all = sps.csr_matrix(self.X_train + self.X_test)
        if item_ids is not None:
            if len(item_ids) != self.X_test.shape[1]:
                raise ValueError("X_train.shape[1] != len(item_ids)")
        self.item_ids = item_ids

    def _X_to_df(self, X: sps.csr_matrix, user_ids: List[Any]) -> pd.DataFrame:
        if self.item_ids is None:
            raise RuntimeError("Setting item_ids is required to use this method.")
        X.sort_indices()
        row, col = X.nonzero()
        data = X.data
        return pd.DataFrame(
            dict(
                user_id=[user_ids[r] for r in row],
                item_id=[self.item_ids[c] for c in col],
                rating=data,
            )
        )

    def df_train(self) -> pd.DataFrame:
        return self._X_to_df(self.X_train, self.user_ids)

    def df_test(self) -> pd.DataFrame:
        return self._X_to_df(self.X_test, self.user_ids)

    def concat(
        self, other: "UserTrainTestInteractionPair"
    ) -> "UserTrainTestInteractionPair":
        """Concatenate the users data.
        user_id will be ``self.user_ids + self.item_ids``.

        Returns:
            [type]: [description]

        ValueError:
            when ``self`` and ``other`` have unequal ``n_items``.
        """
        if self.n_items != other.n_items:
            raise ValueError("inconsistent n_items.")
        return UserTrainTestInteractionPair(
            self.user_ids + other.user_ids,
            sps.vstack([self.X_train, other.X_train], format="csr"),
            sps.vstack([self.X_test, other.X_test], format="csr"),
        )


def split_train_test_userwise_random(
    df_: pd.DataFrame,
    user_colname: str,
    item_colname: str,
    item_ids: List[Any],
    heldout_ratio: float,
    n_heldout: Optional[int],
    rns: np.random.RandomState,
    rating_column: Optional[str] = None,
) -> UserTrainTestInteractionPair:
    """Split the user x item data frame into a pair of sparse matrix (represented as a UserDataSet).

    Parameters
    ----------
    df_:
        user x item interaction matrix.
    user_colname:
        The column name for the users.
    item_colname:
        The column name for the items.
    item_id_to_iid:
        The mapper from item id to item index. If not supplied, create own mapping from df_.
    heldout_ratio:
        The percentage of items (per-user) to be held out as a test(validation) ones.
    n_heldout:
        The maximal number of items (per-user) to be held out as a test(validation) ones.
    rns:
        The random state
    rating_column:
        The column for the rating values. If None, the rating values will be all equal (1), by default None
    Returns
    -------
    UserDataSet
        Resulting train-test split dataset.
    """

    df_ = df_[df_[item_colname].isin(item_ids)]

    item_indices = pd.Categorical(df_[item_colname], categories=item_ids).codes

    user_ids, user_indices = np.unique(df_[user_colname], return_inverse=True)
    if rating_column is not None:
        data = df_[rating_column].values
    else:
        data = np.ones(df_.shape[0], dtype=np.int32)

    X_all = sps.csr_matrix(
        (data, (user_indices, item_indices)),
        shape=(len(user_ids), len(item_ids)),
    )
    X_learn, X_predict = rowwise_train_test_split(
        X_all,
        heldout_ratio,
        n_heldout,
        random_seed=rns.randint(-(2 ** 31), 2 ** 31 - 1),
    )

    return UserTrainTestInteractionPair(
        user_ids, X_learn.tocsr(), X_predict.tocsr(), item_ids
    )


def split_train_test_userwise_time(
    df_: pd.DataFrame,
    user_colname: str,
    item_colname: str,
    time_colname: str,
    item_ids: List[Any],
    heldout_ratio: float,
    n_heldout: Optional[int],
    rating_column: Optional[str] = None,
) -> UserTrainTestInteractionPair:
    df_ = df_[df_[item_colname].isin(item_ids)]
    unique_user_ids, user_index = np.unique(df_[user_colname], return_inverse=True)
    item_index = pd.Categorical(df_[item_colname], categories=item_ids).codes
    df_aux = pd.DataFrame(
        dict(
            user_index=user_index,
            item_index=item_index,
            timestamp=df_[time_colname].values,
        )
    )
    if rating_column is not None:
        df_aux["rating"] = df_[rating_column]
    else:
        df_aux["rating"] = 1.0

    shape_ = (len(unique_user_ids), len(item_ids))

    df_train, df_test = split_last_n_interaction_df(
        df_aux,
        "user_index",
        "timestamp",
        n_heldout=n_heldout,
        heldout_ratio=heldout_ratio,
    )
    X_train = sps.csr_matrix(
        (
            df_train.rating.values,
            (df_train.user_index.values, df_train.item_index.values),
        ),
        shape=shape_,
    )
    X_test = sps.csr_matrix(
        (
            df_test.rating.values,
            (df_test.user_index.values, df_test.item_index.values),
        ),
        shape=shape_,
    )
    return UserTrainTestInteractionPair(unique_user_ids, X_train, X_test, item_ids)


def split_dataframe_partial_user_holdout(
    df_all: pd.DataFrame,
    user_column: str,
    item_column: str,
    time_column: Optional[str] = None,
    rating_column: Optional[str] = None,
    n_val_user: Optional[int] = None,
    n_test_user: Optional[int] = None,
    val_user_ratio: float = 0.1,
    test_user_ratio: float = 0.1,
    heldout_ratio_val: float = 0.5,
    n_heldout_val: Optional[int] = None,
    heldout_ratio_test: float = 0.5,
    n_heldout_test: Optional[int] = None,
    random_state: int = 42,
) -> Tuple[Dict[str, UserTrainTestInteractionPair], List[Any]]:
    """Splits the DataFrame and build an interaction matrix,
    holding out random interactions for a subset of randomly selected users
    (whom we call "validation users" and "test users").

    Args:
        df_all:
            The user-item interaction event log.
        user_column:
            The column name for user_id.
        item_column:
            The column name for movie_id.
        time_column:
            The column name (if any) specifying the time of the interaction.
            If this is set, the split will be based on time, and some of the most recent interactions will be held out for each user.
            Defaults to None.
        rating_column:
            The column name for ratings. If ``None``, the rating will be treated as
            ``1`` for all interactions. Defaults to None.
        n_val_user:
            The number of "validation users". Defaults to None.
        n_test_user:
            The number of "test users". Defaults to None.
        val_user_ratio:
            The percentage of "validation users" with respect to all users.
            Ignored when ``n_val_user`` is set. Defaults to 0.1.
        test_user_ratio:
            The percentage of "test users" with respect to all users.
            Ignored when ``n_text_user`` is set. Defaults to 0.1.
        heldout_ratio_val:
            The percentage of held-out interactions for "validation users".
            Ignored if ``n_heldout_val`` is specified. Defaults to 0.5.
        n_heldout_val:
            The maximal number of held-out interactions for "validation users".
        heldout_ratio_test:
            The percentage of held-out interactions for "test users".
            Ignored if ``n_heldout_test`` is specified. Defaults to 0.5.
        n_heldout_val:
            The maximal number of held-out interactions for "test users".
        random_state:
            The random seed for this procedure. Defaults to 42.

    Raises:
        ValueError: When ``n_val_user + n_test_user`` is greater than the number of total users.

    Returns:
        A tuple consisting of:

            1. A dictionary with ``"train"``, ``"val"``, ``"test"`` as its keys and the
               coressponding dataset as its values.
            2. List of unique item ids (which corresponds to the columns of the datasets).
    """
    assert (test_user_ratio <= 1) and (test_user_ratio >= 0)
    assert (val_user_ratio <= 1) and (val_user_ratio >= 0)

    uids: List[Any] = df_all[user_column].unique()
    n_users_all = len(uids)
    if n_val_user is None:
        n_val_user = int(n_users_all * val_user_ratio)
        val_user_ratio = n_val_user / n_users_all
    else:
        if n_val_user > n_users_all:
            raise ValueError("n_val_user exceeds the number of total users.")

    if n_test_user is None:
        n_test_user = int(n_users_all * test_user_ratio)
        test_user_ratio = n_test_user / n_users_all
    else:
        if (n_test_user + n_val_user) > n_users_all:
            raise ValueError(
                "n_val_user + n_test_users exceeds the number of total users."
            )

    df_all = df_all.drop_duplicates([user_column, item_column])
    rns = np.random.RandomState(random_state)
    train_uids, val_test_uids = train_test_split(
        uids, test_size=(n_val_user + n_test_user), random_state=rns
    )

    if (test_user_ratio * len(uids)) >= 1:
        val_uids, test_uids = train_test_split(
            val_test_uids,
            test_size=n_test_user,
            random_state=rns,
        )
    else:
        val_uids = val_test_uids
        test_uids = np.asarray([], dtype=val_uids.dtype)
    df_train = df_all[df_all[user_column].isin(train_uids)].copy()
    df_val = df_all[df_all[user_column].isin(val_uids)]
    df_test = df_all[df_all[user_column].isin(test_uids)]
    item_all: List[Any] = list(set(df_all[item_column]))

    item_id_to_iid = {id_: i for i, id_ in enumerate(item_all)}
    for df in [df_train, df_val, df_test]:
        df["item_iid"] = df[item_column].map(item_id_to_iid)

    train_uids = df_train[user_column].unique()
    train_uid_to_iid = {uid: iid for iid, uid in enumerate(train_uids)}

    train_user_row = df_train[user_column].map(train_uid_to_iid).values
    train_user_col = df_train.item_iid
    if rating_column is None:
        train_user_data = np.ones(df_train.shape[0])
    else:
        train_user_data = df_train[rating_column].values

    train_user_interactions = sps.csr_matrix(
        (train_user_data, (train_user_row, train_user_col)),
        shape=(len(train_uids), len(item_all)),
    )

    valid_data: Dict[str, UserTrainTestInteractionPair] = dict(
        train=UserTrainTestInteractionPair(train_uids, train_user_interactions, None)
    )
    val_test_info_: List[Tuple[pd.DataFrame, str, float, Optional[int]]] = [
        (df_val, "val", heldout_ratio_val, n_heldout_val),
        (df_test, "test", heldout_ratio_test, n_heldout_test),
    ]
    for df_, dataset_name, heldout_ratio, n_heldout in val_test_info_:
        if time_column is None:
            valid_data[dataset_name] = split_train_test_userwise_random(
                df_,
                user_column,
                item_column,
                item_all,
                heldout_ratio,
                n_heldout,
                rns,
                rating_column=rating_column,
            )
        else:
            valid_data[dataset_name] = split_train_test_userwise_time(
                df_,
                user_column,
                item_column,
                time_column,
                item_all,
                heldout_ratio,
                n_heldout,
                rating_column=rating_column,
            )
    return valid_data, item_all



from typing import Optional

import numpy as np
from scipy import linalg
from scipy import sparse as sps

from irspack.definitions import DenseScoreArray, InteractionMatrix
from irspack.parameter_tuning import CategoricalSuggestion, LogUniformSuggestion
from irspack.user_cold_start.recommenders.base import (
    BaseUserColdStartRecommender,
    ProfileMatrix,
)


def enlarge_profile(X_profile: ProfileMatrix) -> ProfileMatrix:
    X_profile_enlarged = np.zeros(
        (X_profile.shape[0], 1 + X_profile.shape[1]), dtype=np.float32
    )
    X_profile_enlarged[:, 0] = 1
    if sps.issparse(X_profile):
        coo = X_profile.tocoo()
        X_profile_enlarged[coo.row, coo.col + 1] = coo.data
    else:
        X_profile_enlarged[:, 1:] = X_profile
    return X_profile_enlarged


class LinearMethodRecommender(BaseUserColdStartRecommender):
    W: Optional[np.ndarray]

    def __init__(
        self,
        X_interaction: InteractionMatrix,
        X_profile: ProfileMatrix,
        reg: float = 1.0,
        fit_intercept: bool = False,
    ):
        super().__init__(X_interaction, X_profile)
        self.reg = reg
        self.fit_intercept = fit_intercept
        self.W = None

    def _learn(self) -> None:
        if self.fit_intercept:
            X_profile_local = enlarge_profile(self.X_profile)
        else:
            X_profile_local = self.X_profile

        if sps.issparse(X_profile_local):
            X_profile_local = X_profile_local.toarray()
        X_l = X_profile_local.T.dot(X_profile_local)
        index = np.arange(X_l.shape[0])
        X_l[index, index] += self.reg
        inv = np.zeros_like(X_l)
        inv[index, index] = 1
        C_, lower = linalg.cho_factor(X_l, overwrite_a=True)
        inv = linalg.cho_solve((C_, lower), inv, overwrite_b=True)
        self.W = inv.dot(self.X_interaction.T.dot(X_profile_local).T)
        self.W = self.W.reshape(self.W.shape, order="F")

    def get_score(self, profile: ProfileMatrix) -> DenseScoreArray:
        if self.W is None:
            raise RuntimeError("(get_score) called before learning")
        if self.fit_intercept:
            profile = enlarge_profile(profile)
        if sps.issparse(profile):
            profile = profile.toarray()
        return profile.dot(self.W)



from typing import Optional, Tuple

import numpy as np
import pandas as pd


def split_last_n_interaction_df(
    df: pd.DataFrame,
    user_column: str,
    timestamp_column: str,
    n_heldout: Optional[int] = None,
    heldout_ratio: float = 0.1,
) -> Tuple[pd.DataFrame, pd.DataFrame]:
    """Split a dataframe holding out last `n_heldout` or
    last `heldout_ratio` part of interactions of the users.

    Args:
        df:
            The Dataframe to be split.
        user_column :
            The column name for users.
        timestamp_column :
            The column name for "timestamp" (it doesn't have to be datetime).
        n_heldout :
            If not `None`, specifies the maximal number of last actions to be held-out. Defaults to None.
        heldout_ratio :
            Specifies how much of each user interaction will be held out.
            Ignored if ``n_heldout`` is present.

    Returns:
        First interactions and held-out interactions.
    """
    df_sorted = df.sort_values([user_column, timestamp_column])

    index_within_group = df_sorted.groupby(user_column).cumcount(ascending=False).values

    test_indicator: np.ndarray
    if n_heldout is not None:
        test_indicator = index_within_group < n_heldout
    else:
        n_user_appearnce = (
            df_sorted[user_column].value_counts().reindex(df_sorted[user_column].values)
        )
        test_indicator = index_within_group <= (n_user_appearnce * heldout_ratio).values
    return (
        df_sorted.iloc[np.where(~test_indicator)],
        df_sorted.iloc[np.where(test_indicator)],
    )



import logging

from colorlog import ColoredFormatter

IRSPACK_LOGGER_NAME = "IRSPACK"
_logger = logging.getLogger(IRSPACK_LOGGER_NAME)
_logger.propagate = False
_formatter = ColoredFormatter(
    "%(log_color)s[IRSPACK:%(levelname)-1.1s %(asctime)s]%(reset)s %(blue)s%(message)s",
    datefmt=None,
    reset=True,
    log_colors={
        "DEBUG": "cyan",
        "INFO": "green",
        "WARNING": "yellow",
        "ERROR": "red",
        "CRITICAL": "red",
    },
)
_handler = logging.StreamHandler()
_handler.setFormatter(_formatter)
_logger.addHandler(_handler)
_logger.setLevel(logging.INFO)


def get_default_logger() -> logging.Logger:
    return _logger


def disable_default_handler() -> None:
    _logger.removeHandler(_handler)


__all__ = ["get_default_logger", "disable_default_handler"]



# If extensions (or modules to document with autodoc) are in another directory,
# add these directories to sys.path here. If the directory is relative to the
# documentation root, use os.path.abspath to make it absolute, like shown here.
#
import os
import sys
from typing import List

# Configuration file for the Sphinx documentation builder.
#
# This file only contains a selection of the most common options. For a full
# list see the documentation:
# https://www.sphinx-doc.org/en/master/usage/configuration.html

# -- Path setup --------------------------------------------------------------


# sys.path.insert(0, os.path.abspath('../../'))

# -- Project information -----------------------------------------------------

project = "irspack"
copyright = "2020, Tomoki Ohtsuki"
author = "Tomoki Ohtsuki"

# The full version, including alpha/beta/rc tags
release = "0.1.0"


# -- General configuration ---------------------------------------------------

# Add any Sphinx extension module names here, as strings. They can be
# extensions coming with Sphinx (named 'sphinx.ext.*') or your custom
# ones.
extensions = [
    "nbsphinx",
    "sphinx.ext.autodoc",
    "sphinx.ext.autosummary",
    "sphinx.ext.todo",
    "sphinx.ext.viewcode",
    "sphinx.ext.napoleon",
    "sphinx.ext.mathjax",
    "sphinx_rtd_theme",
]

suppress_warnings = [
    "nbsphinx.localfile",
    "nbsphinx.gallery",
    "nbsphinx.thumbnail",
    "nbsphinx.notebooktitle",
    "nbsphinx.ipywidgets",
]

napoleon_google_docstring = True
napoleon_numpy_docstring = False
napoleon_include_private_with_doc = False
napoleon_include_special_with_doc = False
# napoleon_use_admonition_for_examples = False
# napoleon_use_admonition_for_notes = False
# napoleon_use_admonition_for_references = False
# napoleon_use_ivar = True
# napoleon_use_param = True
# napoleon_use_rtype = True


autosummary_generate = True
autoclass_content = "class"
autodoc_typehints = "description"
autodoc_type_aliases = {
    "InteractionMatrix": "Union[scipy.sparse.csr_matrix, scipy.sparse.csc_matrix]",
    "UserIndexArray": "np.ndarray",
    "DenseScoreArray": "np.ndarray",
}
autodoc_default_options = {
    "inherited-members": True,
    "show-inheritance": True,
}
# Add any paths that contain templates here, relative to this directory.
templates_path = ["_templates"]

# List of patterns, relative to source directory, that match files and
# directories to ignore when looking for source files.
# This pattern also affects html_static_path and html_extra_path.
exclude_patterns: List[str] = ["_build", "**.ipynb_checkpoints"]


# -- Options for HTML output -------------------------------------------------

# The theme to use for HTML and HTML Help pages.  See the documentation for
# a list of builtin themes.
#
html_theme = "sphinx_rtd_theme"

# Add any paths that contain custom static files (such as style sheets) here,
# relative to this directory. They are copied after the builtin static files,
# so a file named "default.css" will overwrite the builtin "default.css".
html_static_path = ["_static"]

master_doc = "index"



from abc import ABCMeta, abstractmethod
from typing import TYPE_CHECKING, Any, Dict, Optional, Union

import numpy as np
from optuna.trial import Trial
from scipy import sparse as sps

if TYPE_CHECKING:
    from .. import evaluator

from ..definitions import (
    DenseMatrix,
    DenseScoreArray,
    InteractionMatrix,
    UserIndexArray,
)


def _sparse_to_array(U: Any) -> np.ndarray:
    if sps.issparse(U):
        return U.toarray()
    else:
        return U


class CallBeforeFitError(Exception):
    pass


class BaseRecommender(object, metaclass=ABCMeta):
    X_train_all: sps.csr_matrix
    """The base class for all (hot) recommenders.

    Args:
        X_train_all (csr_matrix|csc_matrix|np.ndarray): user/item interaction matrix.
            each row correspods to a user's interaction with items.
    """

    def __init__(self, X_train_all: InteractionMatrix, **kwargs: Any) -> None:
        self.X_train_all = sps.csr_matrix(X_train_all).astype(np.float64)
        self.n_users: int = self.X_train_all.shape[0]
        self.n_items: int = self.X_train_all.shape[1]
        self.X_train_all.sort_indices()

        # this will store configurable parameters learnt during the training,
        # e.g., the epoch with the best validation score.
        self.learnt_config: Dict[str, Any] = dict()

    def learn(self) -> "BaseRecommender":
        """Learns and returns itself.

        Returns:
            The model after fitting process.
        """
        self._learn()
        return self

    @abstractmethod
    def _learn(self) -> None:
        raise NotImplementedError("_learn must be implemented.")

    def learn_with_optimizer(
        self, evaluator: Optional["evaluator.Evaluator"], trial: Optional[Trial]
    ) -> None:
        """Learning procedures with early stopping and pruning.

        Args:
            evaluator : The evaluator to measure the score.
            trial : The current optuna trial under the study (if any.)
        """
        # by default, evaluator & trial does not play any role.
        self.learn()

    @abstractmethod
    def get_score(self, user_indices: UserIndexArray) -> DenseScoreArray:
        """Compute the item recommendation score for a subset of users.

        Args:
            user_indices : The index defines the subset of users.

        Returns:
            The item scores. Its shape will be (len(user_indices), self.n_items)
        """
        raise NotImplementedError("get_score must be implemented")  # pragma: no cover

    def get_score_block(self, begin: int, end: int) -> DenseScoreArray:
        """Compute the score for a block of the users.

        Args:
            begin (int): where the evaluated user block begins.
            end (int): where the evaluated user block ends.

        Returns:
            The item scores. Its shape will be (end - begin, self.n_items)
        """
        raise NotImplementedError("get_score_block not implemented!")

    def get_score_remove_seen(self, user_indices: UserIndexArray) -> DenseScoreArray:
        """Compute the item score and mask the item in the training set. Masked items will have the score -inf.

        Args:
            user_indices : Specifies the subset of users.

        Returns:
            The masked item scores. Its shape will be (len(user_indices), self.n_items)
        """
        scores = self.get_score(user_indices)
        if sps.issparse(scores):
            scores = scores.toarray()
        m = self.X_train_all[user_indices].tocsr()
        scores[m.nonzero()] = -np.inf
        if scores.dtype != np.float64:
            scores = scores.astype(np.float64)
        return scores

    def get_score_remove_seen_block(self, begin: int, end: int) -> DenseScoreArray:
        """Compute the score for a block of the users, and mask the items in the training set. Masked items will have the score -inf.

        Args:
            begin (int): where the evaluated user block begins.
            end (int): where the evaluated user block ends.

        Returns:
            The masked item scores. Its shape will be (end - begin, self.n_items)
        """
        scores = _sparse_to_array(self.get_score_block(begin, end))
        m = self.X_train_all[begin:end]
        scores[m.nonzero()] = -np.inf
        if scores.dtype != np.float64:
            scores = scores.astype(np.float64)
        return scores

    def get_score_cold_user(self, X: InteractionMatrix) -> DenseScoreArray:
        """Compute the item recommendation score for unseen users whose profiles are given as another user-item relation matrix.

        Args:
            X : The profile user-item relation matrix for unseen users.
                Its number of rows is arbitrary, but the number of columns must be self.n_items.

        Returns:
            Computed item scores for users. Its shape is equal to X.
        """
        raise NotImplementedError(
            f"get_score_cold_user is not implemented for {self.__class__.__name__}!"
        )  # pragma: no cover

    def get_score_cold_user_remove_seen(self, X: InteractionMatrix) -> DenseScoreArray:
        """Compute the item recommendation score for unseen users whose profiles are given as another user-item relation matrix. The score will then be masked by the input.

        Args:
            X : The profile user-item relation matrix for unseen users.
                Its number of rows is arbitrary, but the number of columns must be self.n_items.

        Returns:
            Computed & masked item scores for users. Its shape is equal to X.
        """
        score = self.get_score_cold_user(X)
        score[X.nonzero()] = -np.inf
        return score


class BaseSimilarityRecommender(BaseRecommender):
    W_: Optional[Union[sps.csr_matrix, sps.csc_matrix, np.ndarray]]

    def __init__(self, *args: Any, **kwargs: Any) -> None:
        super().__init__(*args, **kwargs)
        self.W_ = None

    @property
    def W(self) -> Union[sps.csr_matrix, sps.csc_matrix, np.ndarray]:
        """Computed similarity weight matrix.

        Raises:
            RuntimeError: Raises When there is not W_ attributes (e.g., method call before the fit).

        Returns:
            The similarity matrix. Score will be computed by e.g. self.X.dot(self.W)
        """
        if self.W_ is None:
            raise RuntimeError("W fetched before fit.")
        return self.W_

    def get_score(self, user_indices: UserIndexArray) -> DenseScoreArray:
        return _sparse_to_array(self.X_train_all[user_indices].dot(self.W))

    def get_score_cold_user(self, X: InteractionMatrix) -> DenseScoreArray:
        return _sparse_to_array(X.dot(self.W))

    def get_score_block(self, begin: int, end: int) -> DenseScoreArray:
        return _sparse_to_array(self.X_train_all[begin:end].dot(self.W))


class BaseUserSimilarityRecommender(BaseRecommender):
    U_: Optional[Union[sps.csr_matrix, sps.csc_matrix, np.ndarray]]

    def __init__(self, *args: Any, **kwargs: Any) -> None:
        super().__init__(*args, **kwargs)
        self._X_csc: sps.csc_matrix = self.X_train_all.tocsc()
        self.U_ = None

    @property
    def U(self) -> Union[sps.csr_matrix, sps.csc_matrix, np.ndarray]:
        """Computed similarity weight matrix.

        Raises:
            RuntimeError: Raises When there is not W_ attributes (e.g., method call before the fit).

        Returns:
            The similarity matrix. Score will be computed by e.g. self.X.dot(self.W)
        """
        if self.U_ is None:
            raise RuntimeError("W fetched before fit.")
        return self.U_

    def get_score(self, user_indices: UserIndexArray) -> DenseScoreArray:
        return _sparse_to_array(self.U[user_indices].dot(self._X_csc).toarray())

    def get_score_block(self, begin: int, end: int) -> DenseScoreArray:
        return _sparse_to_array(self.U[begin:end].dot(self._X_csc))


class BaseRecommenderWithUserEmbedding(BaseRecommender):
    """Defines a recommender with user embedding (e.g., matrix factorization.).
    These class can be a base CF estimator for CB2CF (with user profile -> user embedding NN).
    """

    @abstractmethod
    def get_user_embedding(
        self,
    ) -> DenseMatrix:
        """Get user embedding vectors.

        Returns:
            The latent vector representation of users.
            Its number of rows is equal to the number of the users.
        """

    @abstractmethod
    def get_score_from_user_embedding(
        self, user_embedding: DenseMatrix
    ) -> DenseScoreArray:
        """Compute the item score from user embedding. Mainly used for cold-start scenario.

        Args:
            user_embedding : Latent user representation obtained elsewhere.

        Returns:
            DenseScoreArray: The score array. Its shape will be ``(user_embedding.shape[0], self.n_items)``
        """
        raise NotImplementedError("get_score_from_item_embedding must be implemtented.")


class BaseRecommenderWithItemEmbedding(BaseRecommender):
    """Defines a recommender with item embedding (e.g., matrix factorization.).
    These class can be a base CF estimator for CB2CF (with item profile -> item embedding NN).
    """

    @abstractmethod
    def get_item_embedding(
        self,
    ) -> DenseMatrix:
        """Get item embedding vectors.

        Returns:
            The latent vector representation of items.
            Its number of rows is equal to the number of the items.
        """
        raise NotImplementedError(
            "get_item_embedding must be implemented."
        )  # pragma: no cover

    @abstractmethod
    def get_score_from_item_embedding(
        self, user_indices: UserIndexArray, item_embedding: DenseMatrix
    ) -> DenseScoreArray:
        raise NotImplementedError("get_score_from_item_embedding must be implemented.")



from collections import Counter
from typing import List

import numpy as np
import pandas as pd
import pytest

from irspack.utils.encoders import (
    BinningEncoder,
    CategoricalValueEncoder,
    DataFrameEncoder,
    ManyToManyEncoder,
)


def test_categorical() -> None:
    CUTOFF = 2
    values = ["A", "B", "C", "A", "B"]
    cnt = Counter(values)
    valid_items = {key for key, cnt in cnt.items() if cnt >= CUTOFF}
    rare_items = {key for key, cnt in cnt.items() if cnt < CUTOFF}

    cenc = CategoricalValueEncoder[str](values, min_count=2)
    X = cenc.transform_sparse(values)
    assert X.shape[1] == len(valid_items) + 1
    assert X.sum() == len(values)
    rare_index = np.where([(x in rare_items) for x in values])
    not_rare_index = np.where([(x not in rare_items) for x in values])
    assert np.all(X[rare_index].nonzero()[1] == 0)
    assert np.all(X[not_rare_index].nonzero()[1] > 0)

    RNS = np.random.RandomState(42)
    fvalues: List[float] = RNS.randn(100)
    fenc = BinningEncoder(fvalues, n_percentiles=5)
    X = fenc.transform_sparse(np.sort(fvalues))
    assert np.all(np.diff(X.nonzero()[1]) >= 0)
    irregular_vals = np.asfarray([np.nan, -np.inf, np.inf])
    irregular_index = fenc.transform_sparse(irregular_vals).nonzero()[1]
    assert irregular_index[0] == 0  # nan's position should be 0
    assert irregular_index[1] == 1  # -inf
    assert irregular_index[2] == len(fenc) - 1

    tag_enc = ManyToManyEncoder(["tag1", "tag2", "tag3"], normalize=True)
    location_enc = ManyToManyEncoder(["loc1", "loc2"], normalize=False)
    dfenc = DataFrameEncoder().add_column("cat", cenc).add_column("price", fenc)
    dfenc.add_many_to_many(
        "item_id",
        "tag",
        tag_enc,
    )
    dfenc.add_many_to_many("item_id", "location", location_enc)

    X_all = dfenc.transform_sparse(
        pd.DataFrame(
            [
                dict(item_id=1, cat="A", price=100.0),
                dict(item_id=2, cat="B", price=0),
            ]
        ),
        [
            pd.DataFrame(
                [
                    dict(item_id=1, tag="tag1"),
                    dict(item_id=1, tag="tag2"),
                    dict(item_id=2, tag="tag3"),
                    dict(item_id=3, tag="tag3"),
                ]
            ),
            pd.DataFrame(
                [
                    dict(item_id=1, location="loc1"),
                    dict(item_id=1, location="loc1"),
                    dict(item_id=2, location="loc2"),
                ]
            ),
        ],
    )
    assert X_all.shape[1] == sum(dfenc.encoder_shapes)
    np.testing.assert_allclose(
        (
            (X_all.toarray()[:, -len(location_enc) - len(tag_enc) : -len(location_enc)])
            ** 2
        ).sum(axis=1),
        1,
    )
    np.testing.assert_allclose(
        (X_all.toarray()[:, -len(location_enc) :]).sum(axis=1),
        np.asarray([2, 1]),
    )

    X_mtom_empty = dfenc.transform_sparse(
        pd.DataFrame([dict(item_id=1, cat="A", price=100.0)]),
        [
            pd.DataFrame([dict(item_id=3, tag="")]),
            pd.DataFrame([dict(item_id=3, location="")]),
        ],
    )
    assert X_mtom_empty[:, -len(tag_enc) - len(location_enc)].count_nonzero() == 0

    with pytest.raises(ValueError):
        X_invalid_arg = dfenc.transform_sparse(
            pd.DataFrame([dict(item_id=1, cat="A", price=100.0)]),
            [
                pd.DataFrame([dict(item_id=3, tag="")]),
                # pd.DataFrame([dict(item_id=3, location="")]), # no location df
            ],
        )



import os
import random
from typing import Optional, Tuple

import numpy as np
import pandas as pd
import scipy.sparse as sps

from ..definitions import InteractionMatrix
from ._util_cpp import (
    okapi_BM_25_weight,
    remove_diagonal,
    rowwise_train_test_split_by_fixed_n,
    rowwise_train_test_split_by_ratio,
    sparse_mm_threaded,
    tf_idf_weight,
)


def rowwise_train_test_split(
    X: InteractionMatrix,
    test_ratio: float = 0.5,
    n_test: Optional[int] = None,
    ceil_n_test: bool = False,
    random_seed: Optional[int] = None,
) -> Tuple[InteractionMatrix, InteractionMatrix]:
    """Splits the non-zero elements of a sparse matrix into two (train & test interactions).
    For each row, the ratio of non-zero elements that become the test interaction
    is (approximately) constant.

    Args:
        X:
            The source sparse matrix.
        test_ratio:
            The ratio of test interactions for each row.
            That is, for each row, if it contains ``NNZ``-nonzero elements,
            the number of elements entering into the test interaction
            will be ``math.floor(test_ratio * NNZ)``.
            Defaults to 0.5.
        random_seed:
            The random seed. Defaults to None.

    Returns:
        A tuple of train & test interactions, which sum back to the original matrix.
    """
    if random_seed is None:
        random_seed = random.randint(-(2 ** 32), 2 ** 32 - 1)
    original_dtype = X.dtype
    X_double = X.astype(np.float64)
    if n_test is None:
        X_train_double, X_test_double = rowwise_train_test_split_by_ratio(
            X_double, random_seed, test_ratio, ceil_n_test
        )
    else:
        X_train_double, X_test_double = rowwise_train_test_split_by_fixed_n(
            X_double, random_seed, n_test
        )
    return (
        X_train_double.astype(original_dtype),
        X_test_double.astype(original_dtype),
    )


def get_n_threads(n_threads: Optional[int]) -> int:
    if n_threads is not None:
        return n_threads
    else:
        try:
            n_threads_cand = os.environ.get(
                "IRSPACK_NUM_THREADS_DEFAULT", os.cpu_count()
            )
            return int(n_threads_cand or 1)
        except:
            raise ValueError(
                'failed to interpret "IRSPACK_NUM_THREADS_DEFAULT" as an integer.'
            )


def df_to_sparse(
    df: pd.DataFrame,
    user_colname: str,
    item_colname: str,
    rating_colname: Optional[str] = None,
) -> Tuple[sps.csr_matrix, np.ndarray, np.ndarray]:
    row, unique_user_ids = pd.factorize(df[user_colname], sort=True)
    col, unique_item_ids = pd.factorize(df[item_colname], sort=True)
    if rating_colname is None:
        data = np.ones(df.shape[0])
    else:
        data = np.asfarray(df[rating_colname].values)
    return (
        sps.csr_matrix(
            (data, (row, col)), shape=(len(unique_user_ids), len(unique_item_ids))
        ),
        unique_user_ids,
        unique_item_ids,
    )


__all__ = [
    "rowwise_train_test_split",
    "sparse_mm_threaded",
    "okapi_BM_25_weight",
    "tf_idf_weight",
    "remove_diagonal",
]



from ._optimizers import (
    AsymmetricCosineKNNOptimizer,
    AsymmetricCosineUserKNNOptimizer,
    BaseOptimizer,
    BaseOptimizerWithEarlyStopping,
    CosineKNNOptimizer,
    CosineUserKNNOptimizer,
    DenseSLIMOptimizer,
    IALSOptimizer,
    JaccardKNNOptimizer,
    NMFOptimizer,
    P3alphaOptimizer,
    RP3betaOptimizer,
    SLIMOptimizer,
    TopPopOptimizer,
    TruncatedSVDOptimizer,
    TverskyIndexKNNOptimizer,
)

__all__ = [
    "BaseOptimizer",
    "BaseOptimizerWithEarlyStopping",
    "TopPopOptimizer",
    "IALSOptimizer",
    "DenseSLIMOptimizer",
    "SLIMOptimizer",
    "P3alphaOptimizer",
    "RP3betaOptimizer",
    "CosineKNNOptimizer",
    "AsymmetricCosineKNNOptimizer",
    "CosineUserKNNOptimizer",
    "AsymmetricCosineUserKNNOptimizer",
    "JaccardKNNOptimizer",
    "TverskyIndexKNNOptimizer",
    "NMFOptimizer",
    "TruncatedSVDOptimizer",
]

try:
    from ._optimizers import BPRFMOptimizer

    __all__.append("BPRFMOptimizer")
except ImportError:
    pass

try:
    from ._optimizers import MultVAEOptimizer

    __all__.append("MultVAEOptimizer")
except ImportError:
    pass



from collections import OrderedDict
from typing import Dict, List, Optional, Tuple

import pandas as pd
import scipy.sparse as sps

from .base import BaseEncoder
from .multi_value import ManyToManyEncoder


class DataFrameEncoder:
    def __init__(self) -> None:
        self.encoders: Dict[str, BaseEncoder] = OrderedDict()
        self.multivalue_encoders: List[Tuple[str, str, str, ManyToManyEncoder]] = []

    def add_column(self, column: str, mapper: BaseEncoder) -> "DataFrameEncoder":
        self.encoders[column] = mapper
        return self

    @property
    def encoder_shapes(self) -> List[int]:
        return [len(encoder) for encoder in self.encoders.values()] + [
            len(encoder) for _, _, _, encoder in self.multivalue_encoders
        ]

    def add_many_to_many(
        self,
        left_key: str,
        target_column: str,
        encoder: ManyToManyEncoder,
        right_key: Optional[str] = None,
    ) -> "DataFrameEncoder":
        if right_key is None:
            right_key = left_key
        self.multivalue_encoders.append((left_key, target_column, right_key, encoder))
        return self

    def transform_sparse(
        self,
        main_table: pd.DataFrame,
        many_to_many_dfs: List[pd.DataFrame] = [],
    ) -> sps.csr_matrix:

        if len(many_to_many_dfs) != len(self.multivalue_encoders):
            raise ValueError(
                f"You have to supply {len(self.multivalue_encoders)} child datafarames."
            )
        Xs: List[sps.csr_matrix] = []
        for colname, mapper in self.encoders.items():
            Xs.append(mapper.transform_sparse(main_table[colname]))

        for df_, (
            index_name_main,
            item_colname,
            index_name_child,
            encoder,
        ) in zip(many_to_many_dfs, self.multivalue_encoders):
            Xs.append(
                encoder.transform_sparse(
                    main_table,
                    df_,
                    index_name_main=index_name_main,
                    item_column_name=item_colname,
                    index_name_items=index_name_child,
                )
            )
        return sps.hstack(Xs, format="csr")
